/* $NetBSD$ */

/*-
 * Copyright (c) 2009 Jared D. McNeill <jmcneill@invisible.ca>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/dkio.h>
#include <sys/disklabel.h>

#include <dev/pud/pud_msgif.h>
#include <dev/pud/pud_disk.h>

#include <prop/proplib.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "libvdisk.h"

#ifndef USE_ARG
#define USE_ARG(x)  /*LINTED*/(void)&x
#endif

/* from kern/subr_disk.c */
static u_int
dkcksum(struct disklabel *d)
{
	size_t npartitions = d->d_npartitions;
	u_short *start, *end;
	u_short sum = 0;

	start = (u_short *)(void *)d;
	end = (u_short *)(void *)&d->d_partitions[npartitions];
	while (start < end)
		sum ^= *start++;
	return sum;
}

static int
vdisk_ioctl_gdiskinfo(struct vdisk_part *vp, void *addr, int flag)
{
	USE_ARG(vp);
	USE_ARG(addr);
	USE_ARG(flag);
	return ENOTSUP;
}

static int
vdisk_ioctl_gdinfo(struct vdisk_part *vp, void *addr, int flag)
{
	struct vdisk *vd = vp->vp_vd;

	USE_ARG(flag);
	memcpy(addr, &vd->vd_label, sizeof(struct disklabel));

	return 0;
}

static int
vdisk_ioctl_gpart(struct vdisk_part *vp, void *addr, int flag)
{
	struct vdisk *vd = vp->vp_vd;
	struct pudpartinfo *pinfo = addr;

	USE_ARG(flag);
	pinfo->disklab = vd->vd_label;
	pinfo->part = vp->vp_part;

	return 0;
}

static int
vdisk_ioctl_sdinfo(struct vdisk_part *vp, void *addr, int flag)
{
	struct vdisk *vd = vp->vp_vd;
	struct disklabel *d = addr;

	if ((flag & FWRITE) == 0)
		return EBADF;

	vd->vd_label = *d;

	return 0;
}

static int
vdisk_ioctl_wdinfo(struct vdisk_part *vp, void *addr, int flag)
{
	int error;

	error = vdisk_ioctl_sdinfo(vp, addr, flag);
	if (error)
		return error;

	return 0;	/* XXX */
}

static int
vdisk_ioctl_gdeflabel(struct vdisk_part *vp, void *addr, int flag)
{
	struct vdisk *vd = vp->vp_vd;
	struct vdisk_if *vif = &vd->vd_if;
	struct disklabel *d = addr;
	uint64_t devsz;

	USE_ARG(flag);

	vif->sizedev(vd->vd_cookie, &devsz);

	memset(d, 0, sizeof(*d));
	d->d_secsize = DEV_BSIZE;
	if (btodb(devsz) >= (32 * 64)) {
		d->d_nsectors = 32;
		d->d_ntracks = 64;
		d->d_ncylinders = (unsigned)btodb(devsz) / (64 * 32);
	} else {
		d->d_nsectors = 1;
		d->d_ntracks = 1;
		d->d_ncylinders = (unsigned)btodb(devsz);
	}
	d->d_secpercyl = d->d_ntracks * d->d_nsectors;
	d->d_type = DTYPE_LD;
	strlcpy(d->d_typename, "unknown", sizeof(d->d_typename));
	strlcpy(d->d_packname, "fictitious", sizeof(d->d_typename));
	d->d_secperunit = (unsigned)btodb(devsz);
	d->d_rpm = 7200;
	d->d_interleave = 0;
	d->d_flags = 0;

#if 1
	d->d_partitions[0].p_offset = 0;
	d->d_partitions[0].p_size =
	    d->d_secperunit * (d->d_secsize / DEV_BSIZE);
	d->d_partitions[0].p_fstype = FS_BSDFFS;
#endif

	d->d_partitions[RAW_PART].p_offset = 0;
	d->d_partitions[RAW_PART].p_size =
	    d->d_secperunit * (d->d_secsize / DEV_BSIZE);
	d->d_partitions[RAW_PART].p_fstype = FS_UNUSED;
	d->d_npartitions = RAW_PART + 1;

	d->d_magic = DISKMAGIC;
	d->d_magic2 = DISKMAGIC;
	d->d_checksum = dkcksum(d);

	return 0;
}

void
vdisk_getdisklabel(struct vdisk *vd)
{
	vdisk_ioctl_gdeflabel(&vd->vd_vp[RAW_PART], &vd->vd_label, 0);
	/* TODO: readdisklabel */
}

int
vdisk_ioctl(struct vdisk_part *vp, u_long cmd, void *addr, int flag)
{
	struct vdisk *vd = vp->vp_vd;
	struct vdisk_if *vif = &vd->vd_if;
	dev_t dev;
	int error;

	if (vif->ioctl) {
		dev = MAKEDISKDEV(VDISK_MAJOR, vd->vd_unit, vp->vp_part);
		error = vif->ioctl(vd->vd_cookie, dev, cmd, addr, flag);
		if (error != EPASSTHROUGH)
			return error;
	}

	switch (cmd) {
	case DIOCGDISKINFO:
		return vdisk_ioctl_gdiskinfo(vp, addr, flag);
	case DIOCGDINFO:
		return vdisk_ioctl_gdinfo(vp, addr, flag);
	case PDIOCGPART:
		return vdisk_ioctl_gpart(vp, addr, flag);
	case DIOCSDINFO:
		return vdisk_ioctl_sdinfo(vp, addr, flag);
	case DIOCWDINFO:
		return vdisk_ioctl_wdinfo(vp, addr, flag);
	case DIOCKLABEL:
	case DIOCWLABEL:
		if ((flag & FWRITE) == 0)
			return EBADF;
		return ENOTSUP;
	case DIOCGDEFLABEL:
		return vdisk_ioctl_gdeflabel(vp, addr, flag);
	case DIOCCACHESYNC:
		if ((flag & FWRITE) == 0)
			return EBADF;
#if notyet
		if (vif->sync)
			return vif->sync(vd->vd_cookie);
#endif
		return ENOTTY;
	default:
		return ENOTTY;
	}
}
